% calcSteeringMatrices : Calculate steerable pyramid steering matrices 
%
% c = calcSteeringMatrices(nbands,alpha)
%
% Inputs:
% "nbands" - number of oriented bands for steering matrices
% "alpha" - vector of N angles for which the steering matrices should be
% computed
%
% Outputs:
% "c" - a collection of steering matrices. "c" will be an 
% [nbands, nbands, N] size array where N is the number of elements of 
% alpha, such that c(:,:,n) is the steering matrix for angle alpha(n).
%
% If b_1, b_2 , ... b_nbands are a set of nband band steerable pyramid
% filters, then for each k=1...nbands,
% c(k,1,n)*b_1 + c(k,2,n)*b_2 + ... + c(k,nbands,n)*b_nbands
% will be the k'th filter, rotated by alpha(n).
%
% Equivalently, if d_1 ... d_nbands are the filter outputs (responses) for
% the nband band steerable filters at some point, then 
% c(k,1,n)*d_1 + ... c(k,nbands,n)*d_nbands is the output of the k'th 
% filter rotated by alpha(n) at the same point

function c = calcSteeringMatrices(nbands, alpha)

for n=1:nbands
  M(:,n) = T( (n-1)*pi/nbands , nbands);
end
iM=inv(M);

c=zeros(nbands,nbands,numel(alpha));

for n = 1:nbands
  c(n,:,:) = iM*T(alpha + (n-1)*pi/nbands , nbands);
end

% T(alpha) will be a vector of coefficients such that
% cos(theta-alpha)^(K-1) = sum_i=1^K T(alpha)_i v_i
%
% where v_i = cos(theta)^(K-i) sin(theta)^(i-1)
% and K is number of orientation bands (K=nbands)
function T = T(alpha,nbands)
T=zeros(nbands,numel(alpha));
ca=cos(alpha(:))';
sa=sin(alpha(:))';
for n=1:nbands
  T(n,:) = ca.^(nbands-n).*sa.^(n-1);
end
