% oagsmncDenoise : perform oagsm/nc denoising for additive white noise
%
% [imd,aux]=oagsmncDenoise(imn,sigma_n,varargin)
%
% This is the gateway routine implementing the denoising algorithm 
% based on the OAGSM/NC model, as described in 
% Hammond and Simoncelli, 2008 (IEEE transaction on Image Processing)
%
% Inputs : 
% imn - noisy image
% sigma_n - standard deviation of noise (assumed to be white)
%
% Other "control parameters" may be set by calling as
% oagsmncDenoise(imn,sigma_n,'param1',value1,'param2',value2)
% These include
% Ntheta - # of samplings to use for theta hidden variable
% Nz - # of samplings to use for z hidden variable
% nbands - # of orientation bands for steerable pyramid transform
% patchrad - controls patch size (patches will be 
% (2*patchrad+1)x(2*patchrad+1) )
% useparent - flag to include parent coefficient in generalized patch
% beta_override - set to impose beta (if set, algorithm will not compute
% beta at each subband by EM). In particular, setting beta_override = 0 
% yields the GSM algorithm, setting beta_override = 1 yields the OAGSM
% algorithm.
% representation - 'fs' for oriented highpass, 's' for normal SP with
% single scalar highpass band.
%
% Other control parameters define lower level aspects of implementation.
%
% Outputs : 
% imd - estimated denoised image
% aux - auxillary structure containing further detail of
%       some parameters calculated during denoising

function [imd,aux]=oagsmncDenoise(imn,sigma_n,varargin)
  control_params={'Ntheta',16,...
                  'Nz',13,... 
                  'nbands',2,... % # of orientation bands for SP
                  'patchrad',2,... % patches are (2*patchrad+1)x(2*patchrad+1)
                  'useparent',1,...
                  'nscales',5,...
                  'beta_override',[],... % set to 0 or 1 to force gsm / oagsm
                  'representation','fs',... % 'fs' for oriented hp bands
                  'Uf',3,...
                  'logzmin',-20.5,...
                  'logzmax',3.5,...                                    
                  'boundary',32,...
                  'beta_emiter_n',20,...
                 }; 
  
  argselectAssign(control_params);
  argselectCheck(control_params,varargin);
  argselectAssign(varargin);
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % set parameters for controlling integration over z and theta
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  logz_list = linspace(logzmin,logzmax,Nz);
  zmean=mean(exp(logz_list));
  theta_list = linspace(-pi/2,pi/2,Ntheta+1);
  theta_list = theta_list(2:end);

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % set generalized neighborhood descriptor
  % K  used when accessing all orientation bands at one scale 
  % K1 used when accessing individual orientation bands separately
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  if ~(useparent)
    K=struct('nbands',nbands,'useparent',useparent,'child',[patchrad patchrad]);
  else
    K=struct('nbands',nbands,'useparent',useparent,...
             'child',[patchrad patchrad],'parent',[0 0],'pfactor',2);
  end
  K1=K;
  K1.nbands=1;

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % control logic for turning on / off rotation
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  if ~isempty(beta_override) && (beta_override == 0)
    do_rotation=0;
  else
    do_rotation=1;
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % boundary handling
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  imn = extendBoundary(imn,boundary,'method','reflect2');
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % build pyramid representation
  % note : we build one more scale than needed 
  % in order to provide parent for coarsest band
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  fprintf('Building pyramid - ');
  switch representation
   case 's'
    [pyr_n,pind]=buildSFpyr(imn,nscales+1,K.nbands-1); 
    lowpass_bandnum = K.nbands*(nscales+1)+2; 
   case 'fs'
    [pyr_n,pind]=buildFullSFpyr2(imn,nscales+1,K.nbands-1); 
    lowpass_bandnum = K.nbands*(nscales+1)+2+K.nbands; 
   otherwise
    error('Unknown representation value : should be s or fs');
  end  
  lowpass_n=pyrBand(pyr_n,pind,lowpass_bandnum);
  % allocate space for denoised result
  pyr_d=zeros(size(pyr_n));

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % build 2-band pyramid for local orientation estimation
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  K2 = struct('nbands',2,'useparent',0,'child',[2 2]);
  [pyr_n2,pind2]=buildSFpyr(imn,nscales,1);
   
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % precompute noise covariances
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
  Khp = struct('useparent',0,'child',K.child,'nbands',1);
  fprintf('Calculating noise covariances');
  [Cnhp,Cn]=calcNoiseCovariance(sigma_n,size(imn),K,representation,Khp);
  fprintf('\n');

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % process bands, highpass first, then loop over bandpass bands
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  % denoise highpass band (with GSM)
  highpass_n_joint = fullPyrBand(pyr_n,pind,K,1);
  switch representation
   case 'fs'
    num_hp_bands = K.nbands;
   case 's'
    num_hp_bands=1;
  end
  
  for b=1:num_hp_bands
    switch representation
     case 'fs'
      hp_bandnum = 1+b;
     case 's'
      hp_bandnum = 1;  
    end
    band_hp_n=getBandFromFullBand(highpass_n_joint,b);
    interior_mask = makeInteriorMask(size(band_hp_n.child),K.child(1)); 
    % calculate hp signal covariance
    allp=packVecPatches(band_hp_n,K1);
    ip=allp(:,interior_mask(:));
    Chp=ip*ip'/size(ip,2);
    Chp = fixNegEigs(Chp-Cnhp{b});
    
    fprintf('Denoising highpass band (%g) \n',b);
    % passing 0 for beta makes oagsmncBand do GSM denoising
    highpass_d = oagsmncBand(band_hp_n,Cnhp{b},[],Chp,logz_list,K1,0);
    pyr_d(pyrBandIndices(pind,hp_bandnum))=highpass_d;
  end
  
  % denoise bandpass bands (with OAGSM/NC)
  
  % If 'fs' representation is used, then adding one to the scale
  % will give the same bands as if 's' representation is used
  % scale_offset allows reuse of code for accessing bands
  scale_offset = (strcmp(representation,'fs')); 
  
  % main loop over scales 
  for kscale=1:nscales
    fprintf('\nDenoising scale %g\n',kscale);
    twoband = fullPyrBand(pyr_n2,pind2,K2,kscale);
    [localori,r,E]=calcNeighborhoodOri(twoband,2);
    fullband = fullPyrBand(pyr_n,pind,K,kscale+scale_offset);
    M=size(fullband.child,1);N=size(fullband.child,2);
    boundary_curr = floor(boundary/2^(kscale-1));
    interior_mask=makeInteriorMask([M N],K.child(1)); 
    fullband_d=zeros(size(fullband.child));      
    Npatches=sum(interior_mask(:));
    % get non rotated patches (norpatch)
    allp=packVecPatches(fullband,K);
    norpatch=allp(:,interior_mask(:));
    % compute non-oriented covariance
    Cnor0=norpatch*norpatch'/size(norpatch,2);
    Cnor0=fixNegEigs(Cnor0-Cn{kscale});
    
    if do_rotation
      Cori_x=calcRotatedCovariance(fullband,localori,...
                                   interior_mask,theta_list,K,Uf);
      for ktheta=1:numel(theta_list)
        Cori0{ktheta}=fixNegEigs(Cori_x{ktheta}-Cn{kscale});
      end
    else
      % assign garbage value. Cori0 shouldn't be used w/o rotation.
      Cori0=zeros(size(Cnor0)); 
    end    
    
    for b=1:K.nbands
      % initialize 
      x=packVecPatches(getBandFromFullBand(fullband,b),K1);
      x=x(:,interior_mask(:));
      Cn_b = extractBlockDiag(Cn{kscale},K,b);    
      Cori_b = extractBlockDiag(Cori0,K,b); 
      Cnor_b = extractBlockDiag(Cnor0,K,b);
      if isempty(beta_override)
        [pori_b,pnor_b]=computeProbabilities(x,Cnor_b,Cori_b,Cn_b,logz_list);
        beta=emIterBeta(pori_b,pnor_b,0.5,beta_emiter_n);
      else
        beta=beta_override;
      end
      fprintf('band %g - ',b);
      fb = getBandFromFullBand(fullband,b);
      % denoise 
      fullband_d(:,:,b)=oagsmncBand(fb,Cn_b,Cori_b,Cnor_b,logz_list,K1,beta);
      beta_save(kscale,b)=beta;
      % pack into output band
      ind_curr=pyrBandIndices(pind,1+(kscale+scale_offset-1)*K.nbands+b);
      pyr_d(ind_curr)=vec(fullband_d(:,:,b));
    end % for b=1:K.nbands
  end % for kscale=1:Nscales
  
  % put lowpass and coarsest oriented band back, these were not denoised
  % so are currently zero in pyr_d. Just copy them from noisy pyramid.
  pyr_d(pyrBandIndices(pind,lowpass_bandnum) )=lowpass_n;
  for b=1:K.nbands
    indns = pyrBandIndices(pind,1+K.nbands*(nscales+scale_offset)+b);
    pyr_d(indns)=pyr_n(indns);
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % reconstruct 
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  switch representation
   case 's'
    imd_big = reconSFpyr(pyr_d,pind);
   case 'fs'
    imd_big = reconFullSFpyr2(pyr_d,pind);
  end
  
  % post-process (remove boundary segment)
  imd = removeBoundary(imd_big,boundary);

  % include other information in aux variable
  % aux.pyr_d = pyr_d;
  aux.beta_save=beta_save;
