% packVecPatches : extract vectorized patches from pyramid bands
%
% function patches=packVecPatches(fullband,K)
%
% Inputs:
% fullband : pyramid band as fullband structure
% K : generalized neighborhood descriptor
%
% Outputs:
% patches - d x N array of vectorized patches. d is patch dimension
% (depends on K), N is the total number of extracted overlapping patches.
%
% For 5x5 patches including 1x1 parent, if K.nbands=2 then
% one has d = 25*2 + 1*2 = 52
%
% Layout of patch dimensions is as
%
% child band 1
% child band 2
% ...
% child band n
% parent band 1
% parent band 2
% ...
% parent band n
%
% where each child band will have (2*K.child(1)+1)*(2*K.child(2)+1)
% coefficients
function patches=packVecPatches(fullband,K)
bdx_m='reflect';

[M N nbands] = size(fullband.child);
if nbands~=K.nbands
  error('fullband structure and K inconsistent');
end

child_seg = zeros( (2*K.child(1)+1)*(2*K.child(2)+1)*nbands,M*N);
boundary = max(K.child)+2;
for b=1:nbands
  child_e = extendBoundary(fullband.child(:,:,b),boundary,'method',bdx_m);
  for m =1:2*K.child(1)+1
    for n =1:2*K.child(2)+1
      cb1=boundary + m - (K.child(1)+1);
      cb2=boundary + n - (K.child(2)+1);
      k=(b-1)*(2*K.child(1)+1)*(2*K.child(2)+1)+(n-1)*(2*K.child(1)+1)+m;
      child_seg(k,:)=vec(child_e((1:M)+cb1,(1:N)+cb2));
    end
  end
end

parent_seg=[];
if K.useparent
  [M N nbands] = size(fullband.parent);
  parent_seg = zeros((2*K.parent(1)+1)*(2*K.parent(2)+1)*nbands,M*N);
  boundary = K.pfactor*(max(K.parent)+2);
  for b=1:nbands
    parent_e = extendBoundary(fullband.parent(:,:,b),boundary,'method',bdx_m);
    for m = 1:2*K.parent(1)+1
      for n = 1:2*K.parent(2)+1
        pb1 = boundary + K.pfactor*(m-(K.parent(1)+1));
        pb2 = boundary + K.pfactor*(n-(K.parent(2)+1));
        k=(b-1)*(2*K.parent(1)+1)*(2*K.parent(2)+1)+(n-1)*(2*K.parent(1)+1)+m;
        parent_seg(k,:)= vec(parent_e((1:M)+pb1,(1:N)+pb2));
      end
    end
  end
end

patches=[child_seg;parent_seg];